from biobookshelf.main import *
from biobookshelf import *

pd.options.mode.chained_assignment = None  # default='warn' # to disable worining

# define global read-only variables
dict_acc_to_arr_acc_query, dict_acc_to_arr_datatype_acc_query, dict_acc_to_arr_phi_query, dict_acc_to_arr_psi_query, dict_acc_to_arr_ss8_query, dict_acc_to_arr_structure_id_query, dict_acc_to_arr_acc_target, dict_acc_to_arr_datatype_acc_target, dict_acc_to_arr_phi_target, dict_acc_to_arr_psi_target, dict_acc_to_arr_ss8_target, dict_acc_to_arr_structure_id_target = [ dict( ) ] * 12

def BCell_Calculate_Similarity_Scores_in_Aligned_Sequences( dir_file_input, float_thres_avg_score_blosum_weighted, float_thres_avg_score_blosum, float_thres_rsa_correlation, l_window_size, dir_folder_cressp, dir_folder_pipeline, dir_folder_pipeline_temp, flag_only_use_structural_properties_of_query_proteins ) :
    """
    Calculate_Similarity_Scores_in_Aligned_Sequences
    using structural properties of both query and target proteins
    """
    uuid_process = UUID( ) # uuid of current process
    np.warnings.filterwarnings( 'ignore' )
    
    float_scaling_factor_for_diff_torsion_angle = 20 # scaling of absolute difference of torsion angle (np.abs): 'float_scaling_factor_for_diff_torsion_angle' / np.abs( angle diff. ) - 'float_scaling_factor_for_diff_torsion_angle' / ( max np.abs( angle diff. ) or 180 )
    float_scaling_factor_for_diff_acc = 0.25 

    # read input alignment between query and target proteins
    df_matched = pd.read_csv( dir_file_input, sep = '\t' )

    # prepare Blosum62 score matrix including all non-conventional annotations of amino acids # 2020-10-29 02:15:32 
    # read dict_blosum62 from the tsv file
    df_blosum62 = pd.read_csv( f'{dir_folder_cressp}data/blosum62.tsv.gz', sep = '\t' )
    dict_blosum62 = dict( )
    for aa_0, aa_1, score in df_blosum62.values : # sould be in [ 'aa_0', 'aa_1', 'BLOSUM62_score' ] order
        dict_blosum62[ aa_0, aa_1 ] = score
        
    # list of ss8 classifications
    l_ss8 = [ 'G', 'H', 'I', 'E', 'B', 'T', 'S', 'C' ]
    
    # for each window_size
    for int_window_size in l_window_size :
        ''' initialize the output file '''
        dir_file_output = f"{dir_folder_pipeline_temp}{uuid_process}_window_size_{int_window_size}.tsv.gz"
        file_output = gzip.open( dir_file_output, 'wb' )
        l_col = [ 'window_size', 'id_alignment', 'source', 'query_accession', 'target_accession', 'e_value', 'identity', 'alignment_start', 'alignment_end', 'query_start', 'query_end', 'target_start', 'target_end', 'query_subsequence', 'target_subsequence', 'score_blosum', 'score_blosum_weighted', 'sum_of_weights', 'score_similarity_acc', 'score_similarity_phi', 'score_similarity_psi', 'score_similarity_ss8', 'n_residues_acc', 'n_residues_phi', 'n_residues_psi', 'n_residues_ss8', 'correl_coeffi_acc', 'correl_p_value_acc', 'correl_coeffi_phi', 'correl_p_value_phi', 'correl_coeffi_psi', 'correl_p_value_psi', 'prop_pdb_evidence_query', 'prop_pdb_evidence_target', 'structure_id_query', 'count_structure_id_query', 'structure_id_target', 'count_structure_id_target', 'most_frequent_ss8_query', 'count_most_frequent_ss8_query', 'most_frequent_ss8_target', 'count_most_frequent_ss8_target' ]
        file_output.write( ( '\t'.join( l_col ) + '\n' ).encode( ) ) # write header to the output file
        n_errors = 0
        
        df = df_matched # copy reference to the alignment file
        df = df[ df.query_alignment.apply( len ) >= int_window_size ] # drop entries smaller than the 'int_window_size'
        l_l_value = list( )
        ''' for each alignment between query and target protein '''
        for id_alignment, query_accession, target_accession, query_start, query_end, target_start, target_end, query_alignment, target_alignment, e_value, identity, source in df[ [ 'id_alignment', 'query_accession', 'target_accession', 'query_start', 'query_end', 'target_start', 'target_end', 'query_alignment', 'target_alignment', 'e_value', 'identity', 'source' ] ].values :
            try :
                ''' retrieve structural properties of query and target proteins '''
                arr_acc_query = dict_acc_to_arr_acc_query[ query_accession ]
                arr_acc_datatype_query = dict_acc_to_arr_datatype_acc_query[ query_accession ]
                arr_ss8_query = dict_acc_to_arr_ss8_query[ query_accession ]
                arr_acc_target = dict_acc_to_arr_acc_target[ target_accession ]
                arr_acc_datatype_target = dict_acc_to_arr_datatype_acc_target[ target_accession ]
                arr_ss8_target = dict_acc_to_arr_ss8_target[ target_accession ]

                len_seq_query, len_seq_target, len_alignment = len( arr_acc_query ), len( arr_acc_target ), len( query_alignment ) # retrieve lengths of sequences and alignment
                arr_phi_query = dict_acc_to_arr_phi_query[ query_accession ] if query_accession in dict_acc_to_arr_phi_query else np.full( len_seq_query, np.nan ) # phi and psi data might not be available for all query and target sequences. if the data is not available, use an empty array (array filled with np.nan) as phi and psi data
                arr_psi_query = dict_acc_to_arr_psi_query[ query_accession ] if query_accession in dict_acc_to_arr_psi_query else np.full( len_seq_query, np.nan )
                arr_structure_id_query = dict_acc_to_arr_structure_id_query[ query_accession ] if query_accession in dict_acc_to_arr_structure_id_query else np.full( len_seq_query, np.nan )
                arr_phi_target = dict_acc_to_arr_phi_target[ target_accession ] if target_accession in dict_acc_to_arr_phi_target else np.full( len_seq_target, np.nan )
                arr_psi_target = dict_acc_to_arr_psi_target[ target_accession ] if target_accession in dict_acc_to_arr_psi_target else np.full( len_seq_target, np.nan )
                arr_structure_id_target = dict_acc_to_arr_structure_id_target[ target_accession ] if target_accession in dict_acc_to_arr_structure_id_target else np.full( len_seq_target, np.nan )

                ''' initialize '''
                arr_alignment_score_blosum, arr_alignment_acc_query, arr_alignment_acc_datatype_query, arr_alignment_ss8_query, arr_alignment_phi_query, arr_alignment_psi_query, arr_alignment_acc_target, arr_alignment_acc_datatype_target, arr_alignment_ss8_target, arr_alignment_phi_target, arr_alignment_psi_target = np.full( ( 10 + 1, len_alignment ), np.nan, dtype = float ) # initialize arrays for containing scores of alignment and structural properties of aligned target and query sequences # np.nan represents invalid values.
                arr_alignment_structure_id_query, arr_alignment_structure_id_target = np.full( ( 2, len_alignment ), np.nan, dtype = object )
                arr_alignment_gap = np.zeros( len_alignment, dtype = bool ) # True when the position contain a gap
                index_pos_in_query, index_pos_in_target = query_start - 1, target_start - 1 # 1-based to 0-based coordinates        
                ''' map structural properties of query and target proteins to the alignment '''
                # loop through each position in the alignment
                for index_pos_in_alignment, residue_query, residue_target in zip( np.arange( len_alignment ), query_alignment.upper( ), target_alignment.upper( ) ) : 
                    arr_alignment_score_blosum[ index_pos_in_alignment ] = dict_blosum62[ residue_query, residue_target ]
                    if residue_query == '-' : # ignore gaps in the alignment for retrieving aligned structural properties, except for RSA values and datatypes for RSA values (RSA values in the gaps are used as weights of the gap penalty score)
                        arr_alignment_acc_target[ index_pos_in_alignment ] = arr_acc_target[ index_pos_in_target ]
                        arr_alignment_acc_datatype_target[ index_pos_in_alignment ] = arr_acc_datatype_target[ index_pos_in_target ]
                        arr_alignment_acc_query[ index_pos_in_alignment ] = 1
                        arr_alignment_acc_datatype_query[ index_pos_in_alignment ] = np.nan # put invalid datatype value to a position with gap
                        arr_alignment_gap[ index_pos_in_alignment ] = True
                        index_pos_in_target += 1 
                    elif residue_target == '-' : 
                        arr_alignment_acc_target[ index_pos_in_alignment ] = 1
                        arr_alignment_acc_datatype_target[ index_pos_in_alignment ] = np.nan # put invalid datatype value to a position with gap
                        arr_alignment_acc_query[ index_pos_in_alignment ] = arr_acc_query[ index_pos_in_query ]
                        arr_alignment_acc_datatype_query[ index_pos_in_alignment ] = arr_acc_datatype_query[ index_pos_in_query ]
                        arr_alignment_gap[ index_pos_in_alignment ] = True
                        index_pos_in_query += 1
                    else :
                        arr_alignment_acc_query[ index_pos_in_alignment ] = arr_acc_query[ index_pos_in_query ]
                        arr_alignment_acc_datatype_query[ index_pos_in_alignment ] = arr_acc_datatype_query[ index_pos_in_query ]
                        arr_alignment_ss8_query[ index_pos_in_alignment ] = arr_ss8_query[ index_pos_in_query ]
                        arr_alignment_phi_query[ index_pos_in_alignment ] = arr_phi_query[ index_pos_in_query ]
                        arr_alignment_psi_query[ index_pos_in_alignment ] = arr_psi_query[ index_pos_in_query ]
                        arr_alignment_structure_id_query[ index_pos_in_alignment ] = arr_structure_id_query[ index_pos_in_query ]
                        arr_alignment_acc_target[ index_pos_in_alignment ] = arr_acc_target[ index_pos_in_target ]
                        arr_alignment_acc_datatype_target[ index_pos_in_alignment ] = arr_acc_datatype_target[ index_pos_in_target ]
                        arr_alignment_ss8_target[ index_pos_in_alignment ] = arr_ss8_target[ index_pos_in_target ]
                        arr_alignment_phi_target[ index_pos_in_alignment ] = arr_phi_target[ index_pos_in_target ]
                        arr_alignment_psi_target[ index_pos_in_alignment ] = arr_psi_target[ index_pos_in_target ]
                        arr_alignment_structure_id_target[ index_pos_in_alignment ] = arr_structure_id_target[ index_pos_in_target ]
                        index_pos_in_query += 1
                        index_pos_in_target += 1
                """ predict cross-reactivity """
                arr_alignment_acc_query_scaled = arr_alignment_acc_query * 2 # scale relative accessible surface area values (multiply with 2 and replace values > 1 with 1)
                arr_alignment_acc_target_scaled = arr_alignment_acc_target * 2
                arr_alignment_acc_query_scaled[ arr_alignment_acc_query_scaled > 1 ] = 1
                arr_alignment_acc_target_scaled[ arr_alignment_acc_target_scaled > 1 ] = 1
                arr_alignment_acc_weight = arr_alignment_acc_query_scaled * arr_alignment_acc_target_scaled
                arr_alignment_score_blosum_weighted = arr_alignment_score_blosum * arr_alignment_acc_weight
                mask_gap = np.isnan( arr_alignment_score_blosum_weighted )
                if mask_gap.sum( ) > 1 : 
                    pass
                arr_alignment_acc_query_scaled[ arr_alignment_gap ] = np.nan # reassign RSA value of the positions containing gaps to np.nan (representing invalid value) for calculation of similarity score based on RSA values
                arr_alignment_acc_target_scaled[ arr_alignment_gap ] = np.nan # reassign RSA value of the positions containing gaps to np.nan (representing invalid value) for calculation of similarity score based on RSA values
                arr_alignment_score_similarity_acc = ( 1 + float_scaling_factor_for_diff_acc ) - np.abs( arr_alignment_acc_query_scaled - arr_alignment_acc_target_scaled ) * ( 1 + float_scaling_factor_for_diff_acc ) # calculate rsa similarity scores using scaled rsa values after alignment
                arr_alignment_score_similarity_acc[ arr_alignment_score_similarity_acc > 1 ] = 1 

                arr_alignment_phi_diff = np.abs( arr_alignment_phi_query - arr_alignment_phi_target ) # calculate torsion angle similarity scores using scaled torsion angle values after alignment
                mask = arr_alignment_phi_diff > 180 # max difference of torsion angles should be 180
                arr_alignment_phi_diff[ mask ] = 360 - arr_alignment_phi_diff[ mask ]
                arr_alignment_psi_diff = np.abs( arr_alignment_psi_query - arr_alignment_psi_target )
                mask = arr_alignment_psi_diff > 180
                arr_alignment_psi_diff[ mask ] = 360 - arr_alignment_psi_diff[ mask ]
                float_min_diff_torsion_angle = float_scaling_factor_for_diff_torsion_angle / 2
                arr_alignment_phi_diff[ arr_alignment_phi_diff < float_min_diff_torsion_angle ] = float_min_diff_torsion_angle # replace 0 with 'float_min_diff_torsion_angle' to avoid zero division
                arr_alignment_psi_diff[ arr_alignment_psi_diff < float_min_diff_torsion_angle ] = float_min_diff_torsion_angle # replace 0 with 'float_min_diff_torsion_angle' to avoid zero division
                arr_alignment_score_similarity_phi = float_scaling_factor_for_diff_torsion_angle / arr_alignment_phi_diff - float_scaling_factor_for_diff_torsion_angle / 180 # max difference of torsion angles is 180
                arr_alignment_score_similarity_psi = float_scaling_factor_for_diff_torsion_angle / arr_alignment_psi_diff - float_scaling_factor_for_diff_torsion_angle / 180 # max difference of torsion angles is 180
                arr_alignment_score_similarity_phi[ arr_alignment_score_similarity_phi > 1 ] = 1
                arr_alignment_score_similarity_psi[ arr_alignment_score_similarity_psi > 1 ] = 1 

                arr_alignment_score_similarity_ss8 = ( arr_alignment_ss8_query == arr_alignment_ss8_target ).astype( float ) # calculate ss8 classification similarity scores using ss8 classifications after alignment
                arr_alignment_score_similarity_ss8[ np.isnan( arr_alignment_ss8_query ) | np.isnan( arr_alignment_ss8_target ) ] = np.nan

                n_windows = len_alignment + 1 - int_window_size
                for index_pos_start in range( n_windows ) : 
                    slice_window = slice( index_pos_start, index_pos_start + int_window_size )
                    float_score_blosum_for_window = arr_alignment_score_blosum[ slice_window ].sum( )
                    float_score_blosum_weighted_for_window = arr_alignment_score_blosum_weighted[ slice_window ].sum( )
                    arr_alignment_acc_weight_for_window = arr_alignment_acc_weight[ slice_window ]
                    float_sum_of_acc_weights = arr_alignment_acc_weight_for_window[ ~ np.isnan( arr_alignment_acc_weight_for_window ) ].sum( ) # calculate total sum of weights

                    float_score_similarity_acc_for_window, n_valid_residues_score_similarity_acc_for_window = NUMPY_Calculate_Average( arr_alignment_score_similarity_acc[ slice_window ] )
                    float_score_similarity_phi_for_window, n_valid_residues_score_similarity_phi_for_window = NUMPY_Calculate_Average( arr_alignment_score_similarity_phi[ slice_window ] )
                    float_score_similarity_psi_for_window, n_valid_residues_score_similarity_psi_for_window = NUMPY_Calculate_Average( arr_alignment_score_similarity_psi[ slice_window ] )
                    float_score_similarity_ss8_for_window, n_valid_residues_score_similarity_ss8_for_window = NUMPY_Calculate_Average( arr_alignment_score_similarity_ss8[ slice_window ] )
                    float_correl_coeffi_acc, float_correl_p_value_acc, _ = PEARSONR( arr_alignment_acc_query_scaled[ slice_window ], arr_alignment_acc_target_scaled[ slice_window ] )
                    float_correl_coeffi_phi, float_correl_p_value_phi, _ = PEARSONR( arr_alignment_phi_query[ slice_window ], arr_alignment_phi_target[ slice_window ] )
                    float_correl_coeffi_psi, float_correl_p_value_psi, _ = PEARSONR( arr_alignment_psi_query[ slice_window ], arr_alignment_psi_target[ slice_window ] )

                    arr_alignment_acc_datatype_query_for_window = arr_alignment_acc_datatype_query[ slice_window ]
                    arr_alignment_acc_datatype_target_for_window = arr_alignment_acc_datatype_target[ slice_window ]
                    arr_alignment_acc_datatype_query_for_window_valid = arr_alignment_acc_datatype_query_for_window[ ~ np.isnan( arr_alignment_acc_datatype_query_for_window ) ] # drop invalid datatype values
                    arr_alignment_acc_datatype_target_for_window_valid = arr_alignment_acc_datatype_target_for_window[ ~ np.isnan( arr_alignment_acc_datatype_target_for_window ) ] # drop invalid datatype values
                    float_prop_pdb_evidence_query = ( arr_alignment_acc_datatype_query_for_window_valid == 2 ).sum( ) / len( arr_alignment_acc_datatype_query_for_window_valid ) if len( arr_alignment_acc_datatype_query_for_window_valid ) > 0 else np.nan # calculate proportion of residues supported by PDB structures
                    float_prop_pdb_evidence_target = ( arr_alignment_acc_datatype_target_for_window_valid == 2 ).sum( ) / len( arr_alignment_acc_datatype_target_for_window_valid ) if len( arr_alignment_acc_datatype_target_for_window_valid ) > 0 else np.nan # calculate proportion of residues supported by PDB structures

                    ''' retrieve most frequent structure_id at the aligned region (for query and target proteins) '''
                    structure_id_query, count_structure_id_query = DICTIONARY_Find_Max( COUNTER( arr_alignment_structure_id_query[ slice_window ] ) )
                    structure_id_target, count_structure_id_target = DICTIONARY_Find_Max( COUNTER( arr_alignment_structure_id_target[ slice_window ] ) )
                    if structure_id_query is None and count_structure_id_query is None : structure_id_query, count_structure_id_query = np.nan, np.nan # retrieve structure_id of most frequence PDB structure (either from RCSB PDB or structural model)
                    if structure_id_target is None and count_structure_id_target is None : structure_id_target, count_structure_id_target = np.nan, np.nan # retrieve structure_id of most frequence PDB structure (either from RCSB PDB or structural model)
                    
                    ''' find most frequent secondary structure classification (ss8) at the aligned region (for query and target proteins) '''
                    # find major secondary structure classification for the current window
                    # query protein
                    arr_alignment_ss8_query_for_window = arr_alignment_ss8_query[ slice_window ]
                    arr_alignment_ss8_query_for_window_valid = arr_alignment_ss8_query_for_window[ ~ np.isnan( arr_alignment_ss8_query_for_window ) ].astype( int )
                    id_ss8_most_frequent_query, count_id_ss8_most_frequent_query = DICTIONARY_Find_Max( COUNTER( arr_alignment_ss8_query_for_window_valid, ignore_float = True ) )
                    if id_ss8_most_frequent_query is None and count_id_ss8_most_frequent_query is None :
                        str_ss8_query, id_ss8_most_frequent_query, count_id_ss8_most_frequent_query = np.nan, np.nan, np.nan
                    else :
                        id_ss8_most_frequent_query = int( id_ss8_most_frequent_query )
                        str_ss8_query = l_ss8[ id_ss8_most_frequent_query ]
                    # target protein
                    arr_alignment_ss8_target_for_window = arr_alignment_ss8_target[ slice_window ]
                    arr_alignment_ss8_target_for_window_valid = arr_alignment_ss8_target_for_window[ ~ np.isnan( arr_alignment_ss8_target_for_window ) ].astype( int )
                    id_ss8_most_frequent_target, count_id_ss8_most_frequent_target = DICTIONARY_Find_Max( COUNTER( arr_alignment_ss8_target_for_window_valid, ignore_float = True ) )
                    if id_ss8_most_frequent_target is None and count_id_ss8_most_frequent_target is None :
                        str_ss8_target, id_ss8_most_frequent_target, count_id_ss8_most_frequent_target = np.nan, np.nan, np.nan
                    else :
                        id_ss8_most_frequent_target = int( id_ss8_most_frequent_target )
                        str_ss8_target = l_ss8[ id_ss8_most_frequent_target ]
                    
                    ''' retrieve subsequence of query and target proteins and count gaps '''
                    query_alignment_subsequence = query_alignment[ slice_window ]
                    target_alignment_subsequence = target_alignment[ slice_window ]
                    int_gap_count_subsequence_query, int_gap_count_subsequence_target = query_alignment_subsequence.count( '-' ), target_alignment_subsequence.count( '-' )
                    int_gap_count_before_subsequence_query, int_gap_count_before_subsequence_target = query_alignment[ : index_pos_start ].count( '-' ), target_alignment[ : index_pos_start ].count( '-' )
                    
                    ''' perform filtering using given threshold values '''
                    if float_score_blosum_for_window < float_thres_avg_score_blosum * int_window_size : continue # do not write subsequence whose average blosum score is below a given threshold
                    if float_correl_coeffi_acc < float_thres_rsa_correlation : continue # do not write subsequence whose correlation coefficient of rsa values is below a given threshold
                    if float_score_blosum_weighted_for_window < float_thres_avg_score_blosum_weighted * int_window_size : continue # filter records with low avg alignment score weighted with accessiblities
                        
                    ''' write a record to the output file '''
                    l = [ int_window_size, id_alignment, source, query_accession, target_accession, e_value, identity, index_pos_start + 1, index_pos_start + int_window_size, query_start + index_pos_start - int_gap_count_before_subsequence_query, query_start + index_pos_start + int_window_size - 1 - int_gap_count_subsequence_query - int_gap_count_before_subsequence_query, target_start + index_pos_start - int_gap_count_before_subsequence_target, target_start + index_pos_start + int_window_size - 1 - int_gap_count_subsequence_target - int_gap_count_before_subsequence_target, query_alignment_subsequence, target_alignment_subsequence, float_score_blosum_for_window, float_score_blosum_weighted_for_window, float_sum_of_acc_weights, float_score_similarity_acc_for_window, float_score_similarity_phi_for_window, float_score_similarity_psi_for_window, float_score_similarity_ss8_for_window, n_valid_residues_score_similarity_acc_for_window, n_valid_residues_score_similarity_phi_for_window, n_valid_residues_score_similarity_psi_for_window, n_valid_residues_score_similarity_ss8_for_window, float_correl_coeffi_acc, float_correl_p_value_acc, float_correl_coeffi_phi, float_correl_p_value_phi, float_correl_coeffi_psi, float_correl_p_value_psi, float_prop_pdb_evidence_query, float_prop_pdb_evidence_target, structure_id_query, count_structure_id_query, structure_id_target, count_structure_id_target, str_ss8_query, count_id_ss8_most_frequent_query, str_ss8_target, count_id_ss8_most_frequent_target ]
                    file_output.write( ( '\t'.join( list( map( str, l ) ) ) + '\n' ).encode( ) ) # write a record to the output file
            except Exception as e : # count errors
                n_errors += 1
                print( f"error number {n_errors}", traceback.format_exc( ) )
        file_output.close( ) # close output file
    return uuid_process # return the UUID of the current process 
def BCell_Combine_result_files_for_each_window_size( dir_file_input, dir_folder_pipeline, dir_folder_pipeline_temp ) :
    """
    combine result files for each window_size
    """
    for arr_input in pd.read_csv( dir_file_input, sep = '\t', header = None ).values :
        int_window_size = arr_input[ 0 ]
        l_dir_file = glob.glob( f"{dir_folder_pipeline_temp}*_window_size_{int_window_size}.tsv.gz" ) # retrieve list of output files to combine into a single output file
        dir_file_output_combining = f'{dir_folder_pipeline}b_cell.subsequence__window_size_{int_window_size}.combining.tsv.gz'
        dir_file_output_combining_completed = f'{dir_folder_pipeline}b_cell.subsequence__window_size_{int_window_size}.tsv.gz'
        OS_FILE_Combine_Files_in_order( l_dir_file, dir_file_output_combining, flag_use_header_from_first_file = True, remove_n_lines = 1, delete_input_files = True )
        print( f"combining output files for window size {int_window_size} is completed" )
        os.rename( dir_file_output_combining, dir_file_output_combining_completed ) # rename the file once completed
def Predict_B_cell_cross_reactivity( dir_folder_pipeline = None, dir_folder_pipeline_temp = None, n_threads = 1, l_window_size = [ 30 ], float_thres_e_value = 30, flag_only_use_structural_properties_of_query_proteins = False, float_thres_avg_score_blosum_weighted__b_cell = 0.15, float_thres_avg_score_blosum__b_cell = 0, float_thres_rsa_correlation = 0 ) :
    
    """
    Package settings
    """
    name_package = 'cressp'
    dir_remote = 'https://github.com/ahs2202/cressp/raw/main/cressp/' # remote directory from which datafiles will be downloaded
    dir_folder_cressp = f"{pkg_resources.resource_filename( name_package, '' )}/" # directory of the current installed package

    
    """
    Predict_B_cell_cross_reactivity
    """
    dir_file_matched = f'{dir_folder_pipeline}matched.tsv.gz'
    df_matched = pd.read_csv( dir_file_matched, sep = '\t' ) # read alignments between query and target protein sequences
    df_matched.index.name = 'id_alignment' # retrieve id_alignment (index of df_matched) during retrieving subsequences
    df_matched.reset_index( drop = False, inplace = True ) # add id_alignment column to to the dataframe
    print( f"number of records: {len( df_matched )}" )
    df_matched = df_matched[ df_matched.e_value <= float_thres_e_value ] # drop entries with too low global similarity
#     print( f"number of records after filtering: {len( df_matched )}" )
    
    # update global variables
    global dict_acc_to_arr_acc_query, dict_acc_to_arr_datatype_acc_query, dict_acc_to_arr_phi_query, dict_acc_to_arr_psi_query, dict_acc_to_arr_ss8_query, dict_acc_to_arr_structure_id_query, dict_acc_to_arr_acc_target, dict_acc_to_arr_datatype_acc_target, dict_acc_to_arr_phi_target, dict_acc_to_arr_psi_target, dict_acc_to_arr_ss8_target, dict_acc_to_arr_structure_id_target

    # setting for decoding structural properties
    dict_kw_rsa = dict( ascii_min = 33, ascii_max = 126, l_ascii_to_exclude = [ 62 ], n_char = 2, value_min = 0, value_max = 1 )
    dict_kw_torsion_angle = dict( ascii_min = 33, ascii_max = 126, l_ascii_to_exclude = [ 62 ], n_char = 2, value_min = -180, value_max = 180 )
    dict_kw_ss8 = dict( ascii_min = 33, ascii_max = 41, l_ascii_to_exclude = [ 62 ], n_char = 1, value_min = 0, value_max = 8 )
    dict_kw_datatype = dict( ascii_min = 33, ascii_max = 36, l_ascii_to_exclude = [ 62 ], n_char = 1, value_min = 0, value_max = 3 )

    # load structural property data for query protein sequences
    df_protein_query = PD_Select( pd.read_csv( f'{dir_folder_pipeline}protein_query.tsv.gz', sep = '\t' ), id_protein = df_matched.query_accession.unique( ) ) # subset structural property database with id_proteins in the alignments
    dict_acc_to_arr_acc_query = ASCII_Decode( df_protein_query.set_index( 'id_protein' )[ 'rsa___ascii_encoding_2_characters_from_33_to_126__from_0_to_1' ].dropna( ).to_dict( ), ** dict_kw_rsa )
    dict_acc_to_arr_datatype_acc_query = ASCII_Decode( df_protein_query.set_index( 'id_protein' )[ 'rsa_datatype___ascii_encoding_1_character_from_33_to_36__states_Pred_Model_PDB' ].dropna( ).to_dict( ), ** dict_kw_datatype )
    dict_acc_to_arr_phi_query = ASCII_Decode( df_protein_query.set_index( 'id_protein' )[ 'phi___ascii_encoding_2_characters_from_33_to_126__from_-180_to_180' ].dropna( ).to_dict( ), ** dict_kw_torsion_angle )
    dict_acc_to_arr_psi_query = ASCII_Decode( df_protein_query.set_index( 'id_protein' )[ 'psi___ascii_encoding_2_characters_from_33_to_126__from_-180_to_180' ].dropna( ).to_dict( ), ** dict_kw_torsion_angle )
    dict_acc_to_arr_ss8_query = ASCII_Decode( df_protein_query.set_index( 'id_protein' )[ 'ss8___ascii_encoding_1_character_from_33_to_41__states_G_H_I_E_B_T_S_C' ].dropna( ).to_dict( ), ** dict_kw_ss8 )
    dict_fasta = df_protein_query.set_index( 'id_protein' )[ 'structure_id___redundancy_reduced' ].dropna( ).to_dict( )
    dict_acc_to_arr_structure_id_query = dict( ( acc, Decode_List_of_Strings( dict_fasta[ acc ] ) ) for acc in dict_fasta )
    del dict_fasta, df_protein_query
    
    # load structural property data for target protein sequences
    if not flag_only_use_structural_properties_of_query_proteins :
        df_protein_target = PD_Select( pd.read_csv( f'{dir_folder_pipeline}protein_target.tsv.gz', sep = '\t' ), id_protein = df_matched.target_accession.unique( ) ) # subset structural property database with id_proteins in the alignments
        dict_acc_to_arr_acc_target = ASCII_Decode( df_protein_target.set_index( 'id_protein' )[ 'rsa___ascii_encoding_2_characters_from_33_to_126__from_0_to_1' ].dropna( ).to_dict( ), ** dict_kw_rsa )
        dict_acc_to_arr_datatype_acc_target = ASCII_Decode( df_protein_target.set_index( 'id_protein' )[ 'rsa_datatype___ascii_encoding_1_character_from_33_to_36__states_Pred_Model_PDB' ].dropna( ).to_dict( ), ** dict_kw_datatype )
        dict_acc_to_arr_phi_target = ASCII_Decode( df_protein_target.set_index( 'id_protein' )[ 'phi___ascii_encoding_2_characters_from_33_to_126__from_-180_to_180' ].dropna( ).to_dict( ), ** dict_kw_torsion_angle )
        dict_acc_to_arr_psi_target = ASCII_Decode( df_protein_target.set_index( 'id_protein' )[ 'psi___ascii_encoding_2_characters_from_33_to_126__from_-180_to_180' ].dropna( ).to_dict( ), ** dict_kw_torsion_angle )
        dict_acc_to_arr_ss8_target = ASCII_Decode( df_protein_target.set_index( 'id_protein' )[ 'ss8___ascii_encoding_1_character_from_33_to_41__states_G_H_I_E_B_T_S_C' ].dropna( ).to_dict( ), ** dict_kw_ss8 )
        dict_fasta = df_protein_target.set_index( 'id_protein' )[ 'structure_id___redundancy_reduced' ].dropna( ).to_dict( )
        dict_acc_to_arr_structure_id_target = dict( ( acc, Decode_List_of_Strings( dict_fasta[ acc ] ) ) for acc in dict_fasta )
        del dict_fasta, df_protein_target
    else :
        ''' retrieve target protein sequence exist in the alignments '''
        set_id_protein_target = set( df_matched.target_accession.unique( ) )
        dict_fasta_protein_target = FASTA_Read( f"{dir_folder_pipeline}protein_target.fasta" )
        dict_fasta_protein_target = dict( ( h.split( ' ', 1 )[ 0 ], dict_fasta_protein_target[ h ] ) for h in dict_fasta_protein_target if h.split( ' ', 1 )[ 0 ] in set_id_protein_target ) # retrieve sequence_id by spliting the header at the first space (to makes sequence_id consistent with that used with blastp) # subset protein sequences with id_proteins in the alignments
        dict_arr_float_empty = dict( ( h, np.full( len( dict_fasta_protein_target[ h ] ), np.nan ) ) for h in dict_fasta_protein_target )
        
        ''' initialize structural properties of target proteins with empty arrays '''
        dict_acc_to_arr_acc_target = deepcopy( dict_arr_float_empty )
        dict_acc_to_arr_datatype_acc_target = deepcopy( dict_arr_float_empty )
        dict_acc_to_arr_phi_target = deepcopy( dict_arr_float_empty )
        dict_acc_to_arr_psi_target = deepcopy( dict_arr_float_empty )
        dict_acc_to_arr_ss8_target = deepcopy( dict_arr_float_empty )
        dict_acc_to_arr_structure_id_target = dict( ( h, np.full( len( dict_fasta_protein_target[ h ] ), np.nan, dtype = object ) ) for h in dict_fasta_protein_target )
        del set_id_protein_target, dict_fasta_protein_target, dict_arr_float_empty
    
    """ predict b-cell cross-reactivity """
    l_uuid_process = Multiprocessing( df_matched, BCell_Calculate_Similarity_Scores_in_Aligned_Sequences, n_threads, dir_temp = dir_folder_pipeline_temp, global_arguments = [ float_thres_avg_score_blosum_weighted__b_cell, float_thres_avg_score_blosum__b_cell, float_thres_rsa_correlation, l_window_size, dir_folder_cressp, dir_folder_pipeline, dir_folder_pipeline_temp, flag_only_use_structural_properties_of_query_proteins ] ) # process similarity search result with multiple processes, and collect uuid of the processes
    # combine output files for each window size
    Multiprocessing( l_window_size, BCell_Combine_result_files_for_each_window_size, n_threads = min( len( l_window_size ), n_threads ), dir_temp = dir_folder_pipeline_temp, global_arguments = [ dir_folder_pipeline, dir_folder_pipeline_temp ] ) # combine result files for each window_size
    

    #     # Bin similarity scores by acc_query and a given binning size for analysis
    #     size_window_binning = 100
    #     size_overlap_binning = 50

    #     Multiprocessing( l_window_size, Bin_Similarity_Scores, n_threads = min( len( l_window_size ), int( OS_Memory( )[ 'MemAvailable' ] / 1e7 ) ), dir_temp = dir_folder_pipeline_temp ) # combine result files for each window_size
    
    
# def Bin_Similarity_Scores( dir_file_input ) :
#     """
#     # 2021-03-31 13:01:07 
#     bin similarity scores 
#     """
#     for arr_input in pd.read_csv( dir_file_input, header = None, sep = '\t' ).values :
#         int_window_size = int( arr_input[ 0 ] )
#         df_matched = pd.read_csv( dir_file_matched, sep = '\t', low_memory = False )
#         df_subsequence = pd.read_csv( f'{dir_folder_pipeline}subsequence__window_size_{int_window_size}__struc_data_human_only.tsv.gz', sep = '\t', low_memory = False )
#         df_subsequence.index.name = 'id_subsequence'
#         df_subsequence.reset_index( drop = False, inplace = True )
#         df = df_subsequence[ [ 'id_subsequence', "query_start", "query_end", 'score_blosum_weighted' ] ]
#         m = MAP.Map( df_matched.query_accession.to_dict( ) )
#         df[ 'acc_query' ] = df_subsequence.id_alignment.apply( m.a2b )
        
#         # Bin similarity scores by acc_query for analysis
#         s = df[ [ 'score_blosum_weighted', 'acc_query' ] ].groupby( 'acc_query' ).max( ).score_blosum_weighted
#         dir_file_output = f"{dir_folder_pipeline}similarity_score__window_size_{int_window_size}.max.bin_accession.tsv.gz"
#         s.to_csv( dir_file_output, sep = '\t' )
        
#         # Binning by 100 amino acid long region of acc_query with 50 amino acid overlaps
#         # build interval trees for score binning
#         df_uniprot_humans = pd.read_csv( dir_file_human_protein_db, sep = '\t' ) # load human proteins used for sequence similarity search
#         dict_acc_to_it = dict( )
#         for acc, length in df_uniprot_humans[ [ 'id_protein', 'seq_length' ] ].values :
#             dict_acc_to_it[ acc ] = intervaltree.IntervalTree( )
#             start = 0
#             while True :
#                 dict_acc_to_it[ acc ].addi( start, start + size_window_binning, [ f"{acc}_from_{start + 1}_to_{min( start + size_window_binning, length )}" ] )
#                 if start + size_window_binning >= length :  
#                     break
#                 else :
#                     start += size_window_binning - size_overlap_binning
#         # retrieve max_score and id_subsequence with the max_score for each id_feature (interval)
#         dict_id_feature_to_max_score = dict( )
#         dict_id_feature_to_id_subsequence_with_max_score = dict( )
#         for id_subsequence, query_start, query_end, score_blosum_weighted, acc_query in df[ [ 'id_subsequence', "query_start", "query_end", 'score_blosum_weighted', 'acc_query' ] ].values :
#             if acc_query not in dict_acc_to_it : continue 
#             it = dict_acc_to_it[ acc_query ]
#             id_feature = list( it.at( query_start - 1 ).intersection( it.at( query_end - 1 ) ) )[ 0 ][ 2 ][ 0 ]
#             if id_feature not in dict_id_feature_to_max_score :
#                 dict_id_feature_to_max_score[ id_feature ] = score_blosum_weighted
#                 dict_id_feature_to_id_subsequence_with_max_score[ id_feature ] = id_subsequence
#             else :
#                 if dict_id_feature_to_max_score[ id_feature ] < score_blosum_weighted : # update id_subsequence with max score
#                     dict_id_feature_to_max_score[ id_feature ] = score_blosum_weighted
#                     dict_id_feature_to_id_subsequence_with_max_score[ id_feature ] = id_subsequence
#         s_max_score = pd.Series( dict_id_feature_to_max_score )
#         s_max_score.index.name = 'id_feature'
#         s_max_score.name = "score_blosum_weighted"
#         s_id_subsequence_with_max_score = pd.Series( dict_id_feature_to_id_subsequence_with_max_score )
#         s_id_subsequence_with_max_score.index.name = 'id_feature'
#         s_id_subsequence_with_max_score.name = "id_subsequence"
        
#         dir_file_output_binning_by_acc_query_region__score = f"{dir_folder_pipeline}similarity_score__window_size_{int_window_size}.max.bin_{size_window_binning}_overlap_{size_overlap_binning}.score.tsv.gz"
#         dir_file_output_binning_by_acc_query_region__id = f"{dir_folder_pipeline}similarity_score__window_size_{int_window_size}.max.bin_{size_window_binning}_overlap_{size_overlap_binning}.id.tsv.gz"
#         s_max_score.to_csv( dir_file_output_binning_by_acc_query_region__score, sep = '\t' )
#         s_id_subsequence_with_max_score.to_csv( dir_file_output_binning_by_acc_query_region__id, sep = '\t' )


    


#!/usr/bin/env python


from biobookshelf.main import *
from biobookshelf import *


from cressp.structural_property_estimation import Estimate_structural_property
from cressp.cross_reactivity_prediction import Predict_T_cell_cross_reactivity, Predict_B_cell_cross_reactivity
from cressp.web_application import Prepare_data_for_web_application

import argparse
import traceback
import os, sys, getopt
from io import StringIO
import time
import math

pd.options.mode.chained_assignment = None  # default='warn' # to disable worining
import warnings
warnings.filterwarnings( action = 'ignore' )
# from scarab.quality_control import Estimate_structural_property

# retrieve a function for logging
Info = multiprocessing.get_logger( ).info

#     # read dict_blosum62 from the tsv file
#     df_blosum62 = pd.read_csv( f'{dir_folder_cressp}data/blosum62.tsv.gz', sep = '\t' )
#     dict_blosum62 = dict( )
#     for aa_0, aa_1, score in df_blosum62.values : # sould be in [ 'aa_0', 'aa_1', 'BLOSUM62_score' ] order
#         dict_blosum62[ aa_0, aa_1 ] = score

def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', dir_folder_output = 'default', n_threads = 1, l_window_size = [ 30 ], float_thres_e_value = 30, flag_use_HMM_search = False, dir_file_query_hmmdb = 'human', flag_use_rcsb_pdb_only = False, int_number_of_proteins_in_a_batch_during_dnn_prediction = 1000, flag_only_use_structural_properties_of_query_proteins = False, float_thres_avg_score_blosum_weighted__b_cell = 0.15, float_thres_avg_score_blosum__b_cell = 0.0, float_thres_rsa_correlation = 0.0, float_thres_avg_blosum62_score_for_mhc = 2, float_thres_min_mhc_allele_frequency = 0.5, float_thres_binding_affinities_in_nM = 500, flag_replace_unconventional_acid_code = False, flag_use_all_gpu_devices = False, flag_deduplicate_based_on_aligned_subsequences_for_visualization = False ) :
    """
    The main function of Cross-Reactive-Epitope-Search-using-Structural-Properties-of-proteins (CRESSP)
    
    
    
    dir_file_protein_target = None : (Required) an input FASTA file containing target protein sequences.
            
    dir_file_protein_query = 'human' : (Default: UniProt human proteins) an input FASTA file containing query protein sequences.
            
    dir_folder_output = 'default' : (Default: a subdirectory 'cressp_out/' of the current directory) an output directory.
            
    n_threads = 1 : (Default: 1) Number of logical CPUs (threads) to use in the current compute node.
            
    l_window_size = [ 30 ] : list of window sizes for searching cross-reactive epitopes from the alignments between query and target proteins.
            
    float_thres_e_value = 30 : (Default: 30, meaning no filtering) threshold for the global alignment e-value.
            
    flag_use_HMM_search = False : (Default: False) Set this flag to perform HMM search in addition to BLASTP search. HMM profile search is performed with HMMER3. The search usually takes several hours for metagenome-assembled genomes.
            
    dir_file_query_hmmdb = 'human' : (Default: a HMM profile database of 1012 human proteins searched against UniProt Pan Proteomes. These proteins consist of experimentally validated human autoantigens) a file containing HMM DB of query proteins aligned against pan-proteomes.
            
    flag_use_rcsb_pdb_only = False : (Default: False) When calculating consensus structural properties of input proteins, do not use homology-based modeled structures from SWISS-MODEL repositories and only use experimental protein structures from RCSB PDB.
    
    int_number_of_proteins_in_a_batch_during_dnn_prediction = 1000 : (Default: 1000) number of proteins in a batch when predicting structural properties of protein residues not covered by known/predict structures.
            
    flag_only_use_structural_properties_of_query_proteins = False : (Default: False) Only use estimated structural properties of the query proteins when predicting cross-reactivity between target and query proteins (skip the estimation of structural properties of target proteins). When 'dir_file_protein_query' == 'human' (default value), it will significantly reduce computation time by skipping estimation and prediction steps of structural properties of target and query proteins.
            
    float_thres_avg_score_blosum_weighted__b_cell = 0.15 : (Default: 0.15) threshold for average weighted BLOSOM62 alignment score for filtering predicted cross-reactive epitopes.
            
    float_thres_avg_score_blosum__b_cell = 0.0 : (Default: 0) threshold for average BLOSOM62 alignment score for filtering predicted cross-reactive epitopes.
            
    float_thres_rsa_correlation = 0 : (Default: 0) threshold for correlation coefficient of Relative Surface Area (RSA) values for filtering predicted cross-reactive epitopes.
    
    float_thres_avg_blosum62_score_for_mhc = 2.0 : (Default: 2.0) threshold for average BLOSOM62 alignment score for filtering predicted cross-reactive T-cell epitopes (cross-reactive MHC epitopes)
    
    float_thres_min_mhc_allele_frequency = 0.5 : (Default: 0.5) a threshold for filtering out MHC alleles with low allele frequencies. MHC alleles with allele frequency above the threshold for at least one population will be used for cross-reactive T-cell epitope prediction
            
    float_thres_binding_affinities_in_nM = 500 : (Default: 500) a threshold predicted IC50 values for filtering predicted T-cell cross-reactive epitopes. A pair of peptides whose geometric average of predicted binding affinities (IC50) values above this threshold will be removed.
    
    flag_deduplicate_based_on_aligned_subsequences_for_visualization = False : (Default: False) perform the deduplication step for removing redundant predicted cross-reactive epitopes based on aligned subsequences
            
    """
    
    """
    Package settings
    """
    name_package = 'cressp'
    dir_remote = 'https://github.com/ahs2202/cressp/raw/main/cressp/' # remote directory from which datafiles will be downloaded
    dir_folder_cressp = f"{pkg_resources.resource_filename( name_package, '' )}/" # directory of the current installed package


    """
    Parse arguments
    """
    # define logging levels
    logger = multiprocessing.log_to_stderr( )
    logger.setLevel( logging.INFO )
    Info( 'CRESSP: comparative analysis of two sets of proteomes for searching potentially cross-reactive B-cell and T-cell epitopes' )
    Info( '[Pipeline Start] Pipeline Started at ' + TIME_GET_timestamp( True ) )
    
    # fixed arguments
    float_search_thres_e_value = 30 # set e-value threshold for search (fixed for maximum sensitivity)

    
    ''' check whether the program is called from the command-line interface or from an interactive Python programming environment '''
    str_name_program = sys.argv[ 0 ]
    if '/' in str_name_program :
        str_name_program = str_name_program.rsplit( '/', 1 )[ 1 ]
    flag_usage_from_command_line_interface = str_name_program[ : len( 'cressp' ) ] == 'cressp'
    
    ''' parse arguments when the function was called from the command-line interface '''
    if flag_usage_from_command_line_interface :
        # command line arguments
        parser = argparse.ArgumentParser( description = "Cross-Reactive-Epitope-Search-using-Structural-Properties-of-proteins (cressp): a program to find cross-reactive epitopes with structural information from known protein structures." )
        parser.add_argument( "-t", "--dir_file_protein_target", help = "(Required) an input FASTA file containing target protein sequences." )
        parser.add_argument( "-q", "--dir_file_protein_query", help = "(Default: UniProt human proteins) an input FASTA file containing query protein sequences.", default = 'human' )
        parser.add_argument( "-o", "--dir_folder_output", help = "(Default: a subdirectory 'cressp_out/' of the current directory) an output directory", default = "default" )
        parser.add_argument( "-c", "--cpu", help = "(Default: 1) Number of logical CPUs (threads) to use in the current compute node.", default = '1' )
        parser.add_argument( "-w", "--window_size", help = "(Default: 30) list of window sizes separated by comma. Example: 15,30,45", default = "30" )
        parser.add_argument( "-e", "--float_thres_e_value", help = "(Default: 30, meaning no filtering) threshold for the global alignment e-value in a scientific notation Example: 1e-3", default = "30" )
        parser.add_argument( "-H", "--flag_use_HMM_search", help = "(Default: False) Set this flag to perform HMM search in addition to BLASTP search. HMM profile search is performed with HMMER3. The search usually takes several hours for metagenome-assembled genomes", action = 'store_true' )
        parser.add_argument( "-d", "--dir_file_query_hmmdb", help = "(Default: a HMM profile database of 1012 human proteins searched against UniProt Pan Proteomes. These proteins consist of experimentally validated human autoantigens) a file containing HMM DB of query proteins aligned against pan-proteomes", default = "human" )
        parser.add_argument( "-R", "--flag_use_rcsb_pdb_only", help = "(Default: False) When calculating consensus structural properties of input proteins, do not use homology-based modeled structures from SWISS-MODEL repositories and only use experimental protein structures from RCSB PDB", action = 'store_true' )
        parser.add_argument( "-B", "--int_number_of_proteins_in_a_batch_during_dnn_prediction", help = "(Default: 1000) number of proteins in a batch when predicting structural properties of protein residues not covered by known/predict structures.", default = '1000' )
        parser.add_argument( "-Q", "--flag_only_use_structural_properties_of_query_proteins", help = "(Default: False) Only use estimated structural properties of the query proteins when predicting cross-reactivity between target and query proteins (skip the estimation of structural properties of target proteins). When 'dir_file_protein_query' == 'human' (default value), it will significantly reduce computation time by skipping estimation and prediction steps of structural properties of target and query proteins.", action = 'store_true' )
        # for filtering predicted cross-reactive epitopes
        parser.add_argument( "-s", "--float_thres_avg_score_blosum_weighted__b_cell", help = "(Default: 0.15) threshold for average rsa-weighted BLOSOM62 alignment score for filtering predicted cross-reactive b-cell epitopes", default = '0.15' )
        parser.add_argument( "-S", "--float_thres_avg_score_blosum__b_cell", help = "(Default: 0) threshold for average BLOSOM62 alignment score for filtering predicted cross-reactive b-cell epitopes", default = '0.0' )
        parser.add_argument( "-C", "--float_thres_rsa_correlation", help = "(Default: 0) threshold for correlation coefficient of Relative Surface Area (RSA) values for filtering predicted cross-reactive b-cell epitopes", default = '0.0' )
        parser.add_argument( "-b", "--float_thres_avg_blosum62_score_for_mhc", help = "(Default: 2.0) threshold for average BLOSOM62 alignment score for filtering predicted cross-reactive T-cell epitopes (cross-reactive MHC epitopes)", default = '2.0' )
        parser.add_argument( "-m", "--float_thres_min_mhc_allele_frequency", help = "(Default: 0.5) a threshold for filtering out MHC alleles with low allele frequencies. MHC alleles with allele frequency above the threshold for at least one population will be used for cross-reactive T-cell epitope prediction", default = '0.5' )
        parser.add_argument( "-a", "--float_thres_binding_affinities_in_nM", help = "(Default: 500) a threshold predicted IC50 values for filtering predicted T-cell cross-reactive epitopes. A pair of peptides whose geometric average of predicted binding affinities (IC50) values above this threshold will be removed.", default = '500' )
        parser.add_argument( "-U", "--flag_replace_unconventional_acid_code", help = "(Default: False) If this flag is set, unconventional amino acids in the input protein sequences will be replaced with chemically similar amino acid. Specifically, Selenocysteine (U) to Cysteine (C), Pyrrolysine (O) to Tyrosine (Y)", action = 'store_true' )
        parser.add_argument( "-G", "--flag_use_all_gpu_devices", help = "(Default: False) Use all available GPU devices during prediction of RSA values. When this flag is set to True, the RSA prediction might be completed faster, but all GPU memories will be occupied by Tensorflow", action = 'store_true' )
        parser.add_argument( "-D", "--flag_deduplicate_based_on_aligned_subsequences_for_visualization", help = "(Default: False) perform the deduplication step for removing redundant predicted cross-reactive epitopes based on aligned subsequences", action = 'store_true' )
        args = parser.parse_args( )
        if args.dir_file_protein_target is None :
            Info( '[Error] Required argument(s) is missing. to view help message, use -h or --help flag' )
            sys.exit( )

        # parse arguments # no further processing required
        flag_use_HMM_search = args.flag_use_HMM_search
        flag_use_rcsb_pdb_only = args.flag_use_rcsb_pdb_only
        flag_only_use_structural_properties_of_query_proteins = args.flag_only_use_structural_properties_of_query_proteins
        flag_use_all_gpu_devices = args.flag_use_all_gpu_devices
        flag_replace_unconventional_acid_code = args.flag_replace_unconventional_acid_code
        int_number_of_proteins_in_a_batch_during_dnn_prediction = int( args.int_number_of_proteins_in_a_batch_during_dnn_prediction )
        n_threads = min( int( args.cpu ), int( multiprocessing.cpu_count( ) ) ) # max number of n_threads is the CPU number of the current local machine
        float_thres_avg_blosum62_score_for_mhc = float( args.float_thres_avg_blosum62_score_for_mhc )
        float_thres_min_mhc_allele_frequency = float( args.float_thres_min_mhc_allele_frequency )
        l_window_size = list( int( e ) for e in args.window_size.split( ',' ) ) # set default window size
        float_thres_e_value = float( args.float_thres_e_value )
        float_thres_avg_score_blosum_weighted__b_cell = float( args.float_thres_avg_score_blosum_weighted__b_cell ) 
        float_thres_avg_score_blosum__b_cell = float( args.float_thres_avg_score_blosum__b_cell ) 
        float_thres_rsa_correlation = float( args.float_thres_rsa_correlation ) 
        float_thres_binding_affinities_in_nM = float( args.float_thres_binding_affinities_in_nM )

        # parse directory arguments
#         dir_file_protein_target_representative
        dir_file_protein_target = args.dir_file_protein_target
        dir_file_protein_query = args.dir_file_protein_query
        dir_folder_output = args.dir_folder_output
        dir_file_query_hmmdb = args.dir_file_query_hmmdb
    else :
        ''' parse arguments when the function was called from an interactive Python interpreter '''
        if dir_file_protein_target is None :
            Info( "[Error] required input 'dir_file_protein_target' was not given" )
            if flag_usage_from_command_line_interface : sys.exit( )
            else : return - 1

    # handle default settings for input datafiles
    flag_default_protein_query_was_used = False # a flag indicating whether the default protein_query was used
    if dir_file_protein_query == 'human' :
        flag_default_protein_query_was_used = True
        PKG.Download_Data( "data/human/uniprot.tsv.gz", dir_remote, name_package ) # download data
        dir_file_protein_query = f'{dir_folder_cressp}data/human/human_uniprot.fa' # set default 'dir_file_protein_query' # file_name will be used to refer to the given query protein, and thus 'human_uniprot' is used as a file_name.
        if not os.path.exists( dir_file_protein_query ) : # if default human protein fasta file is not available, extract the sequence from the dataframe
            df_protein_query = pd.read_csv( f'{dir_folder_cressp}data/human/uniprot.tsv.gz', sep = '\t' )
            dict_fasta_protein_query = df_protein_query.set_index( 'fasta_header' ).seq.to_dict( )
            FASTA_Write( dir_file_protein_query, dict_fasta = dict_fasta_protein_query )

        if flag_use_HMM_search : 
            if dir_file_query_hmmdb == 'human' :
                PKG.Download_Data( "data/human/hmmdb_autoantigen.hmm.gz", dir_remote, name_package ) # download data
                PKG.Gunzip_Data( "data/human/hmmdb_autoantigen.hmm.gz", name_package ) # unzip data
                dir_file_query_hmmdb = pkg_resources.resource_filename( name_package, 'data/human/hmmdb_autoantigen.hmm' ) # set default 'dir_file_query_hmmdb'
    else :
        if flag_use_HMM_search and dir_file_query_hmmdb == 'human' :
            Info( "[Error] exiting since query proteins other than default human proteins were given, the default HMM profile database cannot be used." )
            if flag_usage_from_command_line_interface : sys.exit( )
            else : return - 1
    # get absolute paths for file arguments        
    dir_file_protein_target = os.path.abspath( dir_file_protein_target )
    dir_file_protein_query = os.path.abspath( dir_file_protein_query )
    dir_file_query_hmmdb = os.path.abspath( dir_file_query_hmmdb )

    # handle default setting for the output folder
    if dir_folder_output == 'default' :
        dir_folder_output = f"{os.getcwd( )}/cressp_out/" # set default output folder
    # get absolute paths for folder arguments        
    dir_folder_output = os.path.abspath( dir_folder_output )

    # handle output folder
    if dir_folder_output[ -1 ] != '/' : # last character of a directory should be '/'
        dir_folder_output += '/'
    # define folder directories for each task
    dir_folder_pipeline = f"{dir_folder_output}pipeline/"
    dir_folder_pipeline_temp = f'{dir_folder_pipeline}temp/' 
    dir_folder_pipeline_struc = f'{dir_folder_pipeline}struc/' # create a working directory of estimating structural properties
    dir_folder_pipeline_web = f'{dir_folder_pipeline}web_application/' # a working directory for exporting data for web applications
    dir_folder_web = f'{dir_folder_output}web_application/'
    # create folders
    for dir_folder in [ dir_folder_output, dir_folder_pipeline, dir_folder_pipeline_temp, dir_folder_pipeline_struc, dir_folder_pipeline_web, dir_folder_web ] :
        os.makedirs( dir_folder, exist_ok = True )


    """
    Read and move input protein Fasta files
    """
    def __Get_File_Name_Without_Extension__( dir_file ) :
        ''' get file_name of a given directory to a fasta file (gzipped or uncompressed) and return it '''
        name_file_without_extension = dir_file.rsplit( '/', 1 )[ 1 ]
        # when a file does not have extension
        if '.' not in name_file_without_extension :
            return name_file_without_extension
        name_file_without_extension, str_file_extension = name_file_without_extension.rsplit( '.', 1 )
        # when a file is gzipped, perform extension spliting one more time
        if str_file_extension.lower( ) == 'gz' :
            name_file_without_extension, str_file_extension = name_file_without_extension.rsplit( '.', 1 )
        return name_file_without_extension
    
    # retrieve file names
    name_file_protein_query = __Get_File_Name_Without_Extension__( dir_file_protein_query ) # retrieve file name containing the query proteins
    name_file_protein_target = __Get_File_Name_Without_Extension__( dir_file_protein_target ) # retrieve file name containing the query proteins
    
    """ check flag """
    dir_file_flag = f"{dir_folder_pipeline}copying_input_files_completed.flag"
    if not os.path.exists( dir_file_flag ) :
        
        try :
            dict_fasta_protein_query = FASTA_Read( dir_file_protein_query )
            FASTA_Write( f"{dir_folder_pipeline}protein_query.fasta", dict_fasta = dict_fasta_protein_query )
            dir_file_protein_query = f"{dir_folder_pipeline}protein_query.fasta" # set directory of fasta file to the new file location
        except :
            print( f"exiting due to an error while reading and moving 'dir_file_protein_query' {dir_file_protein_query}" )
            if flag_usage_from_command_line_interface : sys.exit( )
            else : return - 1

        try :
            dict_fasta_protein_target = FASTA_Read( dir_file_protein_target )
            FASTA_Write( f"{dir_folder_pipeline}protein_target.fasta", dict_fasta = dict_fasta_protein_target )
            dir_file_protein_target = f"{dir_folder_pipeline}protein_target.fasta" # set directory of fasta file to the new file location
        except :
            print( f"exiting due to an error while reading and moving 'dir_file_protein_target' {dir_file_protein_target}" )
            if flag_usage_from_command_line_interface : sys.exit( )
            else : return - 1
            
        """ set flag """
        with open( dir_file_flag, 'w' ) as newfile :
            newfile.write( 'completed\n' )
    else :
        # define directories of input protein sequences and load query protein sequences
        dir_file_protein_query = f"{dir_folder_pipeline}protein_query.fasta" # set directory of fasta file to the new file location
        dir_file_protein_target = f"{dir_folder_pipeline}protein_target.fasta" # set directory of fasta file to the new file location
        dict_fasta_protein_query = FASTA_Read( dir_file_protein_query ) # load protein sequences of query (required during BLASTP output processing)
    Info( "[Task Completion] copying input protein sequence files was completed at " + TIME_GET_timestamp( True ) )
        
    """
    Report external and internal program settings
    """
    # compile setting (metadata)
    dict_cressp_setting = {  
            # program setting
        'n_threads' : n_threads,
        'flag_use_HMM_search' : flag_use_HMM_search,
        'flag_use_rcsb_pdb_only' : flag_use_rcsb_pdb_only,
        'flag_only_use_structural_properties_of_query_proteins' : flag_only_use_structural_properties_of_query_proteins,
        'flag_deduplicate_based_on_aligned_subsequences_for_visualization' : flag_deduplicate_based_on_aligned_subsequences_for_visualization,
        'float_thres_avg_blosum62_score_for_mhc' : float_thres_avg_blosum62_score_for_mhc,
        'float_thres_min_mhc_allele_frequency' : float_thres_min_mhc_allele_frequency,
        'l_window_size' : l_window_size,
        'float_thres_e_value' : float_thres_e_value,
        'float_thres_avg_score_blosum_weighted__b_cell' : float_thres_avg_score_blosum_weighted__b_cell,
        'float_thres_avg_score_blosum__b_cell' : float_thres_avg_score_blosum__b_cell,
        'float_thres_rsa_correlation' : float_thres_rsa_correlation,
        'float_thres_binding_affinities_in_nM' : float_thres_binding_affinities_in_nM,
        'dir_file_protein_target' : dir_file_protein_target,
        'dir_file_protein_query' : dir_file_protein_query,
        'dir_folder_output' : dir_folder_output,
        'dir_file_query_hmmdb' : dir_file_query_hmmdb,
            # internal setting
        'dir_folder_pipeline' : dir_folder_pipeline,
        'dir_folder_pipeline_temp' : dir_folder_pipeline_temp,
        'dir_folder_pipeline_struc' : dir_folder_pipeline_struc,
        'dir_folder_pipeline_web' : dir_folder_pipeline_web,
        'dir_folder_web' : dir_folder_web,
        'name_file_protein_query' : name_file_protein_query,
        'name_file_protein_target' : name_file_protein_target }
    
    Info( f"[Setting] cressp will be run with the following setting: {str( dict_cressp_setting )}" )
    
    ''' export CRESSP setting '''
    dir_file_json_setting_cressp = f"{dir_folder_pipeline}cressp_setting.json"
    if os.path.exists( dir_file_json_setting_cressp ) :
        with open( dir_file_json_setting_cressp, 'r' ) as file :
            j = json.load( file )
        if j != dict_cressp_setting :
            Info( f"[Warning] the current CRESSP setting is different from the previous CRESSP setting recorded in the pipeline folder. The previous setting will be used." )
            with open( dir_file_json_setting_cressp, 'r' ) as file :
                dict_cressp_setting = json.load( file ) # override current CRESSP setting with previous CRESSP setting
    with open( dir_file_json_setting_cressp, 'w' ) as newfile :
        json.dump( dict_cressp_setting, newfile )    
    
    """
    Perform BLASTP alignment
    """
    """ check flag """
    dir_file_flag = f"{dir_folder_pipeline}blastp_completed.flag"
    if not os.path.exists( dir_file_flag ) :

        # create blastp_db using query_protein sequences
        dir_prefix_blastdb_protein_query = f"{dir_folder_pipeline}makeblastdb_out/protein_query"
        os.makedirs( f"{dir_folder_pipeline}makeblastdb_out/", exist_ok = True )  
        OS_Run( [ "makeblastdb", "-in", f"{dir_folder_pipeline}protein_query.fasta", '-dbtype', 'prot', '-parse_seqids', '-max_file_sz', '1GB', '-out', dir_prefix_blastdb_protein_query ], dir_file_stdout = f"{dir_prefix_blastdb_protein_query}.makeblastdb.stdout.txt", dir_file_stderr = f"{dir_prefix_blastdb_protein_query}.makeblastdb.stderr.txt", return_output = False ) # make blast db for protein_query

        # run blastp
        dir_file_blastp_output = f'{dir_folder_pipeline}blastp.tsv'
        OS_Run( [ 'blastp', '-query', dir_file_protein_target, '-db', dir_prefix_blastdb_protein_query, '-out', dir_file_blastp_output, '-outfmt', '6 qaccver saccver pident length mismatch gapopen qstart qend sstart send evalue bitscore btop', '-num_threads', f'{n_threads}', '-evalue', f'{float_search_thres_e_value}' ], dir_file_stdout = f"{dir_folder_pipeline}blastp.stdout.txt", dir_file_stderr = f"{dir_folder_pipeline}blastp.stderr.txt", return_output = False ) # run blastp
        OS_Run( [ 'gzip', dir_file_blastp_output ], dir_file_stdout = f"{dir_file_blastp_output}.gzip.stdout.txt", dir_file_stderr = f"{dir_file_blastp_output}.gzip.stderr.txt", return_output = False ) # compress blastp output
        dir_file_blastp_output += '.gz'
        
        """ set flag """
        with open( dir_file_flag, 'w' ) as newfile :
            newfile.write( 'completed\n' )
    else :
        # define blastp output file
        dir_file_blastp_output = f'{dir_folder_pipeline}blastp.tsv.gz'
    Info( "[Task Completion] BLASTP search was completed at " + TIME_GET_timestamp( True ) )
            
    """
    Perform HMMER alignment using a given HMM profile DB (using a couple of HMM profile DBs would be also helpful in the near future)
    """
    # run hmmsearch according to 'flag_use_HMM_search' flag
    if flag_use_HMM_search : 
        """ check flag """
        dir_file_flag = f"{dir_folder_pipeline}hmmsearch_completed.flag"
        dir_file_hmmsearch_output = f'{dir_folder_pipeline}hmmsearch.out'
        if not os.path.exists( dir_file_flag ) :
            OS_Run( [ 'hmmsearch', '-o', dir_file_hmmsearch_output, '--acc', '--notextw', '--cpu', f'{n_threads}', '-E', f'{float_search_thres_e_value}', dir_file_query_hmmdb, dir_file_protein_target ], dir_file_stdout = f"{dir_folder_pipeline}hmmsearch.stdout.txt", dir_file_stderr = f"{dir_folder_pipeline}hmmsearch.stderr.txt", return_output = False ) # run hmmsearch
            """ set flag """
            with open( dir_file_flag, 'w' ) as newfile :
                newfile.write( 'completed\n' )
        Info( "[Task Completion] HMMER search was completed at " + TIME_GET_timestamp( True ) )
    
    """
    Combine BLASTP and HMMSEARCH outputs 
    """
    dir_file_matched = f'{dir_folder_pipeline}matched.tsv.gz'
    
    """ check flag """
    dir_file_flag = f"{dir_file_matched}.write_completed.flag"
    if not os.path.exists( dir_file_flag ) :
        # load blastp result
        dict_qacc_to_seq = FASTA_Read( dir_file_protein_target ) # read query protein sequences
        dict_qacc_to_seq = dict( ( header.split( ' ', 1 )[ 0 ], dict_qacc_to_seq[ header ] ) for header in list( dict_qacc_to_seq ) )
        set_acc_target = set( dict_qacc_to_seq ) # retrieve a set of acc_target
        df_blastp = BLAST_Read( dir_file_blastp_output, dict_qaccver_to_seq = dict_qacc_to_seq ) 
        df_blastp = df_blastp[ [ 'saccver', 'qaccver', 'sstart', 'send', 'qstart', 'qend', 'subject_seq_aligned', 'query_seq_aligned', 'evalue', 'pident' ] ]
        df_blastp.pident = df_blastp.pident / 100
        df_blastp.columns = [ 'query_accession', 'target_accession', 'query_start', 'query_end', 'target_start', 'target_end', 'query_alignment', 'target_alignment', 'e_value', 'identity' ]
        df_blastp[ 'source' ] = 'blastp'

        # load hmmer result according to 'flag_use_HMM_search' flag
        if flag_use_HMM_search : 
            dict_qacc_to_seq = dict( ( header.split( ' ', 1 )[ 0 ], dict_fasta_protein_query[ header ] ) for header in list( dict_fasta_protein_query ) )
            # ignore if records from profiles that do not match input query protein sequences 
            df = PD_Select( HMMER_HMMSEARCH_Read_output( dir_file_hmmsearch_output ), query_accession = set( dict_qacc_to_seq ) ) 
            l_query_alignment = list( ) # replace query consensus sequence with query sequence
            for query_accession, query_alignment, query_start, query_end in df[ [ 'query_accession', 'query_alignment', 'query_start', 'query_end' ] ].values :
                query_seq = dict_qacc_to_seq[ query_accession ][ query_start - 1 : query_end ] # retrive a subsequence of query sequence
                l_subsequence = list( )
                int_start = 0
                for subsequence in query_alignment.split( '.' ) :
                    int_subsequence_length = len( subsequence )
                    l_subsequence.append( query_seq[ int_start : int_start + int_subsequence_length ] )
                    int_start += int_subsequence_length
                l_query_alignment.append( '-'.join( l_subsequence ) )
            df[ 'query_alignment' ] = l_query_alignment
            df[ 'target_alignment' ] = df.target_alignment.str.upper( ) # target alignment string contains amino acids in lower cases when alignment confidence is low, and it is convenient to convert them to upper characters
            df_hmmer = df
            df_hmmer = df_hmmer[ [ 'query_accession', 'target_accession', 'query_start', 'query_end', 'target_start', 'target_end', 'query_alignment', 'target_alignment', 'conditional_Evalue', 'accuracy' ] ] # subset common columns
            df_hmmer.columns = [ 'query_accession', 'target_accession', 'query_start', 'query_end', 'target_start', 'target_end', 'query_alignment', 'target_alignment', 'e_value', 'identity' ] # rename columns
            df_hmmer[ 'source' ] = 'hmmer' 
            set_acc_query = set( header.split( ' ', 1 )[ 0 ] for header in dict_fasta_protein_query ) # retrieve a set of acc_query
            df_hmmer = PD_Select( df_hmmer, query_accession = set_acc_query ) # exclude records with accessions that does not exist in the input query protein sequences

        df_matched = pd.concat( [ df_hmmer, df_blastp ], ignore_index = True ) if flag_use_HMM_search else df_blastp
        df_matched.to_csv( dir_file_matched, sep = '\t', index = False )

        """ set flag """
        with open( dir_file_flag, 'w' ) as file :
            file.write( f"search results were written at {TIME_GET_timestamp( )}" )
    Info( "[Task Completion] Combining HMMER and BLASTP search results was completed at " + TIME_GET_timestamp( True ) )
    
    """
    Estimate structural properties of proteins 
    """
    
    """ check flag """
    dir_file_flag = f"{dir_folder_pipeline}rsa_estimation_completed.flag"
    if not os.path.exists( dir_file_flag ) :
        # use previously calculated structural properties when the default query proteins were used
        if flag_default_protein_query_was_used :
            shutil.copyfile( f'{dir_folder_cressp}data/human/uniprot.tsv.gz', f'{dir_folder_pipeline}protein_query.tsv.gz' ) 
            # use previously calculated alignment between human proteins and RCSB_PDB structures
            PKG.Download_Data( "data/human/uniprot.blastp_rcsb_pdb.with_aligned_seq.filtered.tsv.gz", dir_remote, name_package ) # download data (alignment between human proteins and RCSB_PDB structures)
            shutil.copyfile( f'{dir_folder_cressp}data/human/uniprot.blastp_rcsb_pdb.with_aligned_seq.filtered.tsv.gz', f'{dir_folder_pipeline}struc/protein_query.blastp_rcsb_pdb.with_aligned_seq.filtered.tsv.gz' ) 

        # estimate structural properties
        for name_file in [ 'protein_target' ] if flag_default_protein_query_was_used else [ 'protein_target', 'protein_query' ] : # skip prediction of query proteins if default query proteins are used
            """ check flag_2 """
            dir_file_flag_2 = f"{dir_folder_pipeline}{name_file}.tsv.gz.completed.flag"
            if not os.path.exists( dir_file_flag_2 ) :
                Estimate_structural_property( f'{dir_folder_pipeline}{name_file}.fasta', n_threads, dir_folder_pipeline, dir_folder_pipeline_temp, flag_use_rcsb_pdb_only, int_number_of_proteins_in_a_batch_during_dnn_prediction, flag_use_all_gpu_devices )

                """ set flag_2 """
                with open( dir_file_flag_2, 'w' ) as newfile :
                    newfile.write( 'completed\n' )
                    
        """ set flag """
        with open( dir_file_flag, 'w' ) as newfile :
            newfile.write( 'completed\n' )
    Info( "[Task Completion] Estimation of structural properties of input protein sequences completed at " + TIME_GET_timestamp( True ) )

    """
    Calculate B-cell epitope similarity scores based on structural properties of proteins 
    """
    
    """ check flag """
    dir_file_flag = f"{dir_file_matched}.predicting_B_cell_cross_reactivity_completed.flag"
    if not os.path.exists( dir_file_flag ) :
        
        Predict_B_cell_cross_reactivity( dir_folder_pipeline, dir_folder_pipeline_temp, n_threads, l_window_size, float_thres_e_value, flag_only_use_structural_properties_of_query_proteins, float_thres_avg_score_blosum_weighted__b_cell, float_thres_avg_score_blosum__b_cell, float_thres_rsa_correlation )
        
        """ set flag """
        with open( dir_file_flag, 'w' ) as newfile :
            newfile.write( 'completed\n' )
    Info( "[Task Completion] Prediction of B-cell cross-reactivity was completed at " + TIME_GET_timestamp( True ) )
    
    """ 
    Calculate T-cell epitope similarity scores based on BLOSUM62 scores and predicted binding affinity scores.
    """
    
    """ check flag """
    dir_file_flag = f"{dir_file_matched}.predicting_T_cell_cross_reactivity_completed.flag"
    if not os.path.exists( dir_file_flag ) :
        
        Predict_T_cell_cross_reactivity( dir_folder_pipeline, float_thres_avg_blosum62_score_for_mhc, float_thres_min_mhc_allele_frequency, float_thres_binding_affinities_in_nM, flag_replace_unconventional_acid_code )
        
        """ set flag """
        with open( dir_file_flag, 'w' ) as newfile :
            newfile.write( 'completed\n' )
    Info( "[Task Completion] Prediction of T-cell cross-reactivity was completed at " + TIME_GET_timestamp( True ) )
    
    """
    Further process and export data for visualization using a web application
    """
    
    """ check flag """
    dir_file_flag = f"{dir_file_matched}.prepare_data_for_web_application_completed.flag"
    if not os.path.exists( dir_file_flag ) :

        # combine results of all 'window_size' values
        OS_FILE_Combine_Files_in_order( glob.glob( f"{dir_folder_pipeline}b_cell.subsequence__window_size_*.tsv.gz" ), f"{dir_folder_pipeline}b_cell.subsequence.tsv.gz", remove_n_lines = 1, flag_use_header_from_first_file = True )
        # prepare data for web application using the combined subsequence
        # copy data for web application and encode using base64 encoding, and write metadata
        Prepare_data_for_web_application( f"{dir_folder_pipeline}b_cell.subsequence.tsv.gz", f"{dir_folder_pipeline}t_cell.mhc_binding.tsv.gz", dict_cressp_setting )
        
        """ set flag """
        with open( dir_file_flag, 'w' ) as newfile :
            newfile.write( 'completed\n' )
    Info( "[Task Completion] Exporting data for CRESSP web application was completed at " + TIME_GET_timestamp( True ) )


def Parse_Structural_Properties( dir_file_df_sp, int_datatype = None, name_col_for_identifying_protein = 'id_protein' ) :
    """
    # 2021-05-31 20:29:02 
    parse structural properties with typical parameters for parsing ascii-encoded structural properties and typical column names  
    
    'dir_file_df_sp' : a file directory to TSV file containing structural properties estimated from CRESSP. Alternatively, a dataframe of the TSV file can be given as an input.
    'int_datatype' : used for initializing 'datatype_acc' if 'rsa_datatype___ascii_encoding_1_character_from_33_to_36__states_Pred_Model_PDB' column does not exist
    'name_col_for_identifying_protein' : columne name of the given tabular data for identifying a unique protein. The unique value in the column will be used as a 'key' in the returned dicitonaries
    
    """
    df_sp = pd.read_csv( dir_file_df_sp, sep = '\t' ) if isinstance( dir_file_df_sp, ( str ) ) else dir_file_df_sp # read database of structural properties # if an object that is not a string datatype is given, assumes 'dir_file_df_sp' is a dataframe containing structural properties
    df_sp.set_index( name_col_for_identifying_protein, inplace = True )
    
    dict_kw_rsa = dict( ascii_min = 33, ascii_max = 126, l_ascii_to_exclude = [ 62 ], n_char = 2, value_min = 0, value_max = 1 )
    dict_kw_torsion_angle = dict( ascii_min = 33, ascii_max = 126, l_ascii_to_exclude = [ 62 ], n_char = 2, value_min = -180, value_max = 180 )
    dict_kw_ss8 = dict( ascii_min = 33, ascii_max = 41, l_ascii_to_exclude = [ 62 ], n_char = 1, value_min = 0, value_max = 8 )
    dict_kw_datatype = dict( ascii_min = 33, ascii_max = 36, l_ascii_to_exclude = [ 62 ], n_char = 1, value_min = 0, value_max = 3 )

    
    dict_sp = dict( )
    dict_sp[ 'acc' ] = ASCII_Decode( df_sp.rsa___ascii_encoding_2_characters_from_33_to_126__from_0_to_1.to_dict( ), ** dict_kw_rsa ) # decode mkdssp outputs of RCSB PDB data
    dict_sp[ 'phi' ] = ASCII_Decode( df_sp[ 'phi___ascii_encoding_2_characters_from_33_to_126__from_-180_to_180' ].to_dict( ), ** dict_kw_torsion_angle )
    dict_sp[ 'psi' ] = ASCII_Decode( df_sp[ 'psi___ascii_encoding_2_characters_from_33_to_126__from_-180_to_180' ].to_dict( ), ** dict_kw_torsion_angle )
    dict_sp[ 'ss8' ] = ASCII_Decode( df_sp[ 'ss8___ascii_encoding_1_character_from_33_to_41__states_G_H_I_E_B_T_S_C' ].to_dict( ), ** dict_kw_ss8 )
    if 'rsa_datatype___ascii_encoding_1_character_from_33_to_36__states_Pred_Model_PDB' in df_sp.columns.values :
        dict_sp[ 'datatype_acc' ] = ASCII_Decode( df_sp[ 'rsa_datatype___ascii_encoding_1_character_from_33_to_36__states_Pred_Model_PDB' ].to_dict( ), ** dict_kw_datatype )
    else :
        if int_datatype is not None :
            return -1
            
        # initialization of a dictionary of arrays containing the identifier of the current dataset of 'acc' datatype
        dict_datatype_acc = dict( )
        for h in dict_sp[ 'acc' ] :
            arr = np.zeros_like( dict_sp[ 'acc' ][ h ] ) # default = 0 (predicted structural property)
            arr[ ~ np.isnan( dict_sp[ 'acc' ][ h ] ) ] = int_datatype
            dict_datatype_acc[ h ] = arr
        dict_sp[ 'datatype_acc' ] = dict_datatype_acc
    if 'structure_id___redundancy_reduced' in df_sp.columns.values :
        dict_fasta = df_sp[ 'structure_id___redundancy_reduced' ].dropna( ).to_dict( )
        dict_sp[ 'structure_id' ] = dict( ( acc, Decode_List_of_Strings( dict_fasta[ acc ] ) ) for acc in dict_fasta )
        dict_sp[ 'seq' ] = df_sp[ 'seq' ].to_dict( )
    
    return dict_sp

def Encode_Structural_Properties( dict_sp, name_col_for_identifying_protein = 'id_protein' ) :
    """
    # 2021-05-31 20:28:58 
    Compose a dataframe containing encoded structural properties using given 'dict_sp'

    'dict_sp' : dictionary containing structural properties and protein sequences
    """
    dict_kw_rsa = dict( ascii_min = 33, ascii_max = 126, l_ascii_to_exclude = [ 62 ], n_char = 2, value_min = 0, value_max = 1 )
    dict_kw_torsion_angle = dict( ascii_min = 33, ascii_max = 126, l_ascii_to_exclude = [ 62 ], n_char = 2, value_min = -180, value_max = 180 )
    dict_kw_ss8 = dict( ascii_min = 33, ascii_max = 41, l_ascii_to_exclude = [ 62 ], n_char = 1, value_min = 0, value_max = 8 )
    dict_kw_datatype = dict( ascii_min = 33, ascii_max = 36, l_ascii_to_exclude = [ 62 ], n_char = 1, value_min = 0, value_max = 3 )

    ''' initialize dataframe with protein sequences '''
    df_sp = pd.Series( dict_sp[ 'seq' ], name = 'seq' ).reset_index( ).rename( columns = { 'index' : name_col_for_identifying_protein } ).set_index( name_col_for_identifying_protein )

    # encode combined structural properties into ASCII strings using ASCII encoding
    df_sp[ 'rsa___ascii_encoding_2_characters_from_33_to_126__from_0_to_1' ] = pd.Series( ASCII_Encode( dict_sp[ 'acc' ], ** dict_kw_rsa ) )
    df_sp[ 'phi___ascii_encoding_2_characters_from_33_to_126__from_-180_to_180' ] = pd.Series( ASCII_Encode( dict_sp[ 'phi' ], ** dict_kw_torsion_angle ) )
    df_sp[ 'psi___ascii_encoding_2_characters_from_33_to_126__from_-180_to_180' ] = pd.Series( ASCII_Encode( dict_sp[ 'psi' ], ** dict_kw_torsion_angle ) )
    df_sp[ 'ss8___ascii_encoding_1_character_from_33_to_41__states_G_H_I_E_B_T_S_C' ] = pd.Series( ASCII_Encode( dict_sp[ 'ss8' ], ** dict_kw_ss8 ) )
    df_sp[ 'rsa_datatype___ascii_encoding_1_character_from_33_to_36__states_Pred_Model_PDB' ] = pd.Series( ASCII_Encode( dict_sp[ 'datatype_acc' ], ** dict_kw_datatype ) )
    df_sp[ 'structure_id___redundancy_reduced' ] = pd.Series( dict( ( h, Encode_List_of_Strings( dict_sp[ 'structure_id' ][ h ] ) ) for h in dict_sp[ 'structure_id' ] ) )
    df_sp.reset_index( drop = False, inplace = True )
    return df_sp
    
    
if __name__ == "__main__" :
    cressp( ) # run CRESSP at the top-level environment
    



from biobookshelf.main import *
from biobookshelf import *

pd.options.mode.chained_assignment = None  # default='warn' # to disable worining

def Predict_T_cell_cross_reactivity( dir_folder_pipeline, float_thres_avg_blosum62_score_for_mhc, float_thres_min_mhc_allele_frequency, float_thres_binding_affinities_in_nM, flag_replace_unconventional_acid_code ) :
    """
    Predict_T_cell_cross_reactivity
    """

    """
    Package settings
    """
    name_package = 'cressp'
    dir_remote = 'https://github.com/ahs2202/cressp/raw/main/cressp/' # remote directory from which datafiles will be downloaded
    dir_folder_cressp = f"{pkg_resources.resource_filename( name_package, '' )}/" # directory of the current installed package

    ''' read dict_blosum62 from the tsv file '''
    df_blosum62 = pd.read_csv( f'{dir_folder_cressp}data/blosum62.tsv.gz', sep = '\t' )
    dict_blosum62 = dict( )
    for aa_0, aa_1, score in df_blosum62.values : # sould be in [ 'aa_0', 'aa_1', 'BLOSUM62_score' ] order
        dict_blosum62[ aa_0, aa_1 ] = score

    """ 
    Extract subsequences from the alignments, filter, predict binding affinity, and prepare an output file containing individual records
    """

    # setting
    int_thres_min_length_peptide_mhc_i = 8
    int_thres_max_length_peptide_mhc_i = 15

    ''' define and open an output folder '''
    dir_file_output = f"{dir_folder_pipeline}t_cell.mhc_binding.tsv.gz"
    newfile = gzip.open( dir_file_output, 'wb' )
    l_col = [ 'mhc_allele', 'affinity_query', 'affinity_target', 'source', 'query_accession', 'target_accession', 'e_value', 'score_blosum', 'query_start', 'query_end', 'target_start', 'target_end', 'query_subsequence', 'target_subsequence', 'window_size', 'average_score_blosum', 'query_processing_score', 'target_processing_score', 'mhc_class', 'score_for_sorting' ]
    newfile.write( ( '\t'.join( l_col ) + '\n' ).encode( ) ) # write a header line for the output file

    ''' load MHCFlurry binding affinity predictor '''
    from mhcflurry import Class1PresentationPredictor # load MHCflurry predictor
    predictor = Class1PresentationPredictor.load( ) # load MHCflurry predictor for benchmarking

    ''' retrieve MHC-I alleles with allele frequency larger then the given threshold in at least one population '''
    PKG.Download_Data( "data/mhc_population_allele_frequency.tsv.gz", dir_remote, name_package ) # download data
    PKG.Download_Data( "data/autoimmune_disease_associated_mhc_alleles.tsv.gz", dir_remote, name_package ) # download data
    df_mhc_af = pd.read_csv( f"{dir_folder_cressp}data/mhc_population_allele_frequency.tsv.gz", sep = '\t', index_col = [ 0, 1 ] )
    l_mhc_i_allele = list( PD_Threshold( ( df_mhc_af.loc[ 'I' ] > float_thres_min_mhc_allele_frequency ).sum( axis = 1 ), a = 0 ).index.values ) + list( pd.read_csv( f"{dir_folder_cressp}data/autoimmune_disease_associated_mhc_alleles.tsv.gz", sep = '\t' ).mhc_allele.unique( ) ) # retrieve mhc_i_alleles with allele frequency > 'float_thres_min_mhc_allele_frequency' in at least one population # also retrieve autoimmune-associated alleles
    l_mhc_i_allele = list( set( l_mhc_i_allele ).intersection( predictor.supported_alleles ) ) # retrive valid alleles for MHCflurry

    ''' prepare df_matched before iteration '''
    if 'df_matched' not in locals( ) :
        df_matched = pd.read_csv( f'{dir_folder_pipeline}matched.tsv.gz', sep = '\t' ) # read alignments between query and target protein sequences
    df_matched_for_mhc_i = df_matched[ df_matched.query_alignment.apply( len ) >= int_thres_max_length_peptide_mhc_i ]
    iter_arr_df_matched_for_mhc_i = ( arr for arr in df_matched_for_mhc_i[ [ 'query_accession', 'target_accession', 'query_start', 'query_end', 'target_start', 'target_end', 'query_alignment', 'target_alignment', 'e_value', 'source' ] ].values )

    ''' initialize a bucket '''
    flag_file_end_reached = False
    l_l_value = [ ] # initialize a bucket
    int_max_num_line_before_start_prediction = 50000 # since predicting multiple records in a batch is much faster, records will be collected in a bucket before they are processed together

    while True :
        ''' iterate 'df_matched_for_mhc_i' '''
        try :
            query_accession, target_accession, query_start, query_end, target_start, target_end, query_alignment, target_alignment, e_value, source = next( iter_arr_df_matched_for_mhc_i )
        except StopIteration :
            flag_file_end_reached = True

        ''' extract and filter subsequences '''
        if not flag_file_end_reached :
            try : l_similarity_score = list( dict_blosum62[ ( amino_acid_query, amino_acid_target ) ] for amino_acid_query, amino_acid_target in zip( query_alignment.upper( ), target_alignment.upper( ) ) ) # calculate similarity score for each residue
            except : 
                continue # in case of ambiguous amino acids (e.g. J = I or L, B = D or N), skip the protein itself
            ''' iterate different int_window_size '''
            for int_window_size in np.arange( int_thres_min_length_peptide_mhc_i, int_thres_max_length_peptide_mhc_i + 1 ) :
                n_windows = len( l_similarity_score ) + 1 - int_window_size
                for int_start in range( n_windows ) : 
                    int_similarity_score_for_window = sum( l_similarity_score[ int_start : int_start + int_window_size ] ) 
                    query_alignment_subsequence = query_alignment[ int_start : int_start + int_window_size ]
                    target_alignment_subsequence = target_alignment[ int_start : int_start + int_window_size ]
                    int_gap_count_subsequence_query, int_gap_count_subsequence_target = query_alignment_subsequence.count( '-' ), target_alignment_subsequence.count( '-' )
                    int_gap_count_before_subsequence_query, int_gap_count_before_subsequence_target = query_alignment[ : int_start ].count( '-' ), target_alignment[ : int_start ].count( '-' )
                    # [ 'source', 'query_accession', 'target_accession', 'e_value', 'score_blosum', 'query_start', 'query_end', 'target_start', 'target_end', 'query_subsequence', 'target_subsequence', 'window_size', 'average_score_blosum' ]
                    ''' filter subsequences based on average blosum62 scores '''
                    average_score_blosum = int_similarity_score_for_window / int_window_size # retrieve an average blosum score
                    if average_score_blosum >= float_thres_avg_blosum62_score_for_mhc :
                        l_l_value.append( [ source, query_accession, target_accession, e_value, int_similarity_score_for_window, query_start + int_start - int_gap_count_before_subsequence_query, query_start + int_start + int_window_size - 1 - int_gap_count_subsequence_query - int_gap_count_before_subsequence_query, target_start + int_start - int_gap_count_before_subsequence_target, target_start + int_start + int_window_size - 1 - int_gap_count_subsequence_target - int_gap_count_before_subsequence_target, query_alignment_subsequence, target_alignment_subsequence, int_window_size, average_score_blosum ] )        

        ''' check whether the bucket has been filled, and predict binding affinity of the peptides '''
        if len( l_l_value ) >= int_max_num_line_before_start_prediction or flag_file_end_reached : # detect whether the bucket has been filled
            ''' compose a dataframe from the collected records '''
            df_subsequence_mhc_i = pd.DataFrame( l_l_value, columns = [ 'source', 'query_accession', 'target_accession', 'e_value', 'score_blosum', 'query_start', 'query_end', 'target_start', 'target_end', 'query_subsequence', 'target_subsequence', 'window_size', 'average_score_blosum' ] )
            ''' initialize the next bucket ''' 
            l_l_value = [ ]

            ''' remove subsequences containing invalid amino acids, and retrieve subsequences without gaps '''
            l_aa_invalid = list( "XBJZ" ) if flag_replace_unconventional_acid_code else list( "OBJUZX" ) # define invalid amino acids based on settings 'flag_replace_unconventional_acid_code'
            df_subsequence_mhc_i = PD_Search( df_subsequence_mhc_i, query_subsequence = l_aa_invalid, target_subsequence = l_aa_invalid, is_negative_query = True ) # retrive subset of df_subsequence for MHC_I allele binding prediction # invalid residues 'B' and 'Z' should be removed 
            df_subsequence_mhc_i[ 'query_subsequence_without_gap' ] = list( seq.replace( '-', '' ).replace( 'U', 'C' ).replace( 'O', 'Y' ) for seq in df_subsequence_mhc_i.query_subsequence.values ) if flag_replace_unconventional_acid_code else list( seq.replace( '-', '' ) for seq in df_subsequence_mhc_i.query_subsequence.values ) # remove gap in the aligment and use the sequence as inputs
            df_subsequence_mhc_i[ 'target_subsequence_without_gap' ] = list( seq.replace( '-', '' ).replace( 'U', 'C' ).replace( 'O', 'Y' ) for seq in df_subsequence_mhc_i.target_subsequence.values ) if flag_replace_unconventional_acid_code else list( seq.replace( '-', '' ) for seq in df_subsequence_mhc_i.target_subsequence.values )

            int_min_len_peptide = 5 # lower limit of the length of the peptide for binding affinity prediction
            df_subsequence_mhc_i = df_subsequence_mhc_i[ ( df_subsequence_mhc_i.query_subsequence_without_gap.apply( len ) >= int_min_len_peptide ) & ( df_subsequence_mhc_i.target_subsequence_without_gap.apply( len ) >= int_min_len_peptide ) ] # drop aligned peptides with peptide shorter than the minimum peptide length of prediction algorithm (MHCflurry)

            ''' predict binding affinity of peptides pairs with MHCflurry '''
            df_subsequence_mhc_i.reset_index( drop = True, inplace = True ) # reset index before join operation
            dict_data = dict( )
            for str_allele in l_mhc_i_allele :
                df_predicted_query_subsequence = predictor.predict( peptides = list( df_subsequence_mhc_i.query_subsequence_without_gap.values ), alleles = [ str_allele ], verbose = 0 ) 
                df_predicted_target_subsequence = predictor.predict( peptides = list( df_subsequence_mhc_i.target_subsequence_without_gap.values ), alleles = [ str_allele ], verbose = 0 ) # remove gap in the aligment and use the sequence as inputs
                if 'query_processing_score' not in dict_data : dict_data[ 'query_processing_score' ] = df_predicted_query_subsequence.processing_score.values # retreve processing score only once for query and target peptides
                if 'target_processing_score' not in dict_data : dict_data[ 'target_processing_score' ] = df_predicted_target_subsequence.processing_score.values
                dict_data[ f'query_affinity_{str_allele}' ] = df_predicted_query_subsequence.affinity.values
                dict_data[ f'target_affinity_{str_allele}' ] = df_predicted_target_subsequence.affinity.values
            df_subsequence_mhc_i = df_subsequence_mhc_i.join( pd.DataFrame( dict_data ) ) # combine predicted result

            ''' prepare a dataframe containing individual records '''
            float_thres_product_binding_affinities = float_thres_binding_affinities_in_nM ** 2 # calculate a threshold for the product of binding affinities for cross-reactivity prediction
             # prepare an array of indices
            df_subsequence_mhc_i.reset_index( drop = True, inplace = True ) # reset index before join operation
            arr_index = df_subsequence_mhc_i.index.values
            l = [ ]
            for mhc_i_allele in l_mhc_i_allele :
                for index, affinity_query, affinity_target in zip( arr_index, dict_data[ f'query_affinity_{mhc_i_allele}' ], dict_data[ f'target_affinity_{mhc_i_allele}' ] ) :
                    ''' perform filtering '''
                    if affinity_query * affinity_target <= float_thres_product_binding_affinities : # drop a pair of peptides if the geometric average of predicted binding affinity values (IC50) is above the threshold
                        l.append( [ index, mhc_i_allele, affinity_query, affinity_target ] )    
            df_subsequence_mhc_i.drop( columns = list( c for c in df_subsequence_mhc_i.columns.values if '_affinity_' in c ) + [ 'query_subsequence_without_gap', 'target_subsequence_without_gap' ], inplace = True ) # drop unnecessary columns
            df_mhc_web = pd.DataFrame( l, columns = [ 'index', 'mhc_allele', 'affinity_query', 'affinity_target' ] ).set_index( 'index' ).join( df_subsequence_mhc_i, how = 'left' ) 
            df_mhc_web[ 'mhc_class' ] = 1 # MHC-Class is I
            df_mhc_web[ 'score_for_sorting' ] = df_mhc_web.score_blosum / ( df_mhc_web.affinity_query * df_mhc_web.affinity_target ) ** 0.5 # calculate score for sorting
            df_mhc_web.to_csv( newfile, sep = '\t', index = False, header = False, mode = 'wb' )

            ''' complete the operation if the end of the file has been reached '''
            if flag_file_end_reached :
                break
            break

    newfile.close( )
